#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from synnax.arc.client import Arc, Client
from synnax.arc.payload import (
    Edge,
    Graph,
    GraphNode,
    Handle,
    Key,
    Mode,
    Payload,
    Position,
    Text,
)
from synnax.arc.types import Task, TaskConfig
from synnax.util.deprecation import deprecated_getattr

_DEPRECATED = {
    "ArcTask": "Task",
    "ArcTaskConfig": "TaskConfig",
    "ArcClient": "Client",
    "ArcKey": "Key",
    "ArcMode": "Mode",
    "ArcPayload": "Payload",
}

__getattr__ = deprecated_getattr(__name__, _DEPRECATED, globals())

__all__ = [
    "Arc",
    "Client",
    "Key",
    "Mode",
    "Payload",
    "Task",
    "TaskConfig",
    "Edge",
    "Graph",
    "GraphNode",
    "Handle",
    "Position",
    "Text",
]
